%% estimateIRFs.m
% Estimates effects of monetary shocks on stock market using SVAR-IV

%% Import and preliminaries

varFile = importdata('data/var.csv');
varData = varFile.data;
varNames = varFile.colheaders; 

ivFile = importdata('data/iv.csv');
ivData = ivFile.data;
ivNames = ivFile.colheaders;

iv = 'mp1_tc';
mpInd = 'gs1';
otherVars = {'logcpi','logip','realyield1mo','excessequity1mo','dp_dt'};
relFundReturns = {'cumrelfundreturn_dt','cumrelmret_dt'};
includeRelFund = 0;

yName = 'excessequity1mo';
rName = 'realyield1mo';
rho = 0.9962; % used in Campbell-Shiller decomposition, from Campbell-Ammer 93

varSeriesToPrint = {'1-yr Treasury','CPI','Industrial production','1-mo real rate','1-mo excess return','Dividend/price'};
relFundSeriesToPrint = {{'Relative wealth,','high beta hedge funds'},{'Relative wealth,','high beta mutual funds'}};

lags = 6; % number of lags to include in VAR
lagsIV = 1; % number of lags of IV to include in LP-IV 
 % (following Stock-Watson 18, given Ramey 16 critique)
lagslocal = 6; % number of lags of variables to include in LP-IV
numBoots = 10000; % number of bootstrapped samples
T = 48; % number of periods to be plotted in IRFs (with shock in period 1)
lCI = 5; % percentile of confidence interval lower bound
uCI = 95; % percentile of confidence interval upper bound

signToPlot = -1; % 1 to plot MP tightening; -1 to plot loosening 
makePlots = 1; % indicator of whether or not to make plots
rowsToPlot = 2;
colsToPlot = 3; % make sure rowsToPlot * colsToPlot >= length(varSeries)

mapIVToVar = nan(size(ivData,1),1);
mapIVToVar(1) = find((ivData(1,strcmp('year',ivNames)) == varData(:,strcmp('year',varNames))) & ...
    (ivData(1,strcmp('month',ivNames)) == varData(:,strcmp('month',varNames))));
mapIVToVar(2:size(mapIVToVar,1)) = mapIVToVar(1) + (1:(size(mapIVToVar,1)-1))'; 

if ~includeRelFund
    fileName = 'baseline';
else
    fileName = 'funds';
    otherVars = [otherVars,relFundReturns];
    varSeriesToPrint = [varSeriesToPrint,relFundSeriesToPrint];
    colsToPlot = 4;
    
    origVarData = varData;
    origMapIVToVar = mapIVToVar;
    varData = varData([[mapIVToVar(1)-lags:mapIVToVar(1)-1]';mapIVToVar],:);
    mapIVToVar = mapIVToVar-mapIVToVar(1)+lags+1;    
end
varSeries = [mpInd,otherVars];

%% Initialize diary

if ~exist('output','dir')
    mkdir('output');
end

delete(['output/',fileName,'.txt']); % clear any old diary
diary(['output/',fileName,'.txt']);

%% Run VAR, estimate impulse responses using SVAR-IV, and compute Campbell-Shiller decomposition

numOtherVars = size(otherVars,2);
numVars = numOtherVars+1;
varInds = nan(1,numVars);
varInds(1) = find(strcmp(mpInd,varNames),1);
for j=1:numOtherVars
    varInds(1+j) = find(strcmp(otherVars{j},varNames),1);
end

varDep = varData(lags+1:size(varData,1),varInds);
varIndep = lagmatrix(varData(:,varInds),1:lags);
varIndep = varIndep(lags+1:size(varIndep,1),:);

sy = zeros(1,numVars*lags);
sy(find(strcmp(yName,otherVars),1)+1) = 1;
sr = zeros(1,numVars*lags);
sd = zeros(1,numVars*lags);
sr(find(strcmp(rName,otherVars),1)+1) = 1;

% Run VAR

varBeta = [varIndep,ones(size(varIndep,1),1)]\varDep;
varResid = varDep - [varIndep,ones(size(varIndep,1),1)]*varBeta;
varSigma = (1/(size(ivData,1)-(lags*numVars+1)))*varResid(mapIVToVar-lags,:)'*varResid(mapIVToVar-lags,:);

varSSR = sum(varResid.^2,1);
varSST = sum((varDep-repmat(mean(varDep,1),size(varDep,1),1)).^2,1);
varRsq = 1 - varSSR./varSST;
varAdjRsq = 1 - (1-varRsq)*(size(varIndep,1)-1)/(size(varIndep,1)-size(varIndep,2)-1);

varSigmaFull = kron((1/(size(varResid,1)-numVars*lags-1))*(varResid'*varResid),eye(size(varResid,1),size(varResid,1)));
varX = kron(eye(numVars,numVars),[varIndep,ones(size(varIndep,1),1)]);
varVBeta = ((varX'*varX)^(-1))*(varX'*varSigmaFull*varX)*((varX'*varX)^(-1));

% Stack VAR into a first-order system
% Let A be the matrix of coefficients on non-constants, in order of thisDep 
%  then thisIndep
A = nan(lags*numVars,lags*numVars);
A(1:numVars,:) = varBeta(1:(size(varBeta,1)-1),:)';
A(numVars+1:lags*numVars,:) = [eye((lags-1)*numVars),zeros((lags-1)*numVars,numVars)];

% Generate impulse responses  

[irf,sp,sq] = genIRFIV(T,varSigma,varResid,ivData,ivNames,mapIVToVar,lags,A,otherVars,iv);

% Decompose contemporaneous effects of MP *loosening* on excess returns

uExcessDecompMP = nan(1,4);
 % [unexpected S&P 500 return, -revised expectation of future excess
 % returns, -revised expectation of future real rates, revised
 % expectation of dividends] 
uExcessDecompMP(1) = - irf(1,sy==1) - irf(1,sr==1);
uExcessDecompMP(2) = sy*rho*A*((eye(numVars*lags)-rho*A)^(-1))*[sp;sq;zeros((lags-1)*numVars,1)];
uExcessDecompMP(3) = sr*rho*A*((eye(numVars*lags)-rho*A)^(-1))*[sp;sq;zeros((lags-1)*numVars,1)];
uExcessDecompMP(4) = uExcessDecompMP(:,1) - uExcessDecompMP(:,2) - uExcessDecompMP(:,3);

%% Add 'lags' values of lagged IV to VAR and report F-stats for hypothesis that they equal zero (a test of invertibility)

Fstats_invert = nan(numVars);
Fstats_pval_invert = nan(numVars);

ivIndep = ivData(:,find(strcmp(iv,ivNames),1));
varIndep_iv = lagmatrix(ivIndep,1:lags);
varIndep_iv = [varIndep_iv(lags+1:length(ivIndep),:),varIndep(mapIVToVar(1+lags:length(mapIVToVar))-lags,:)];
varDep_iv = varDep(mapIVToVar(1+lags:length(mapIVToVar))-lags,:);
varBeta_iv = [varIndep_iv,ones(size(varIndep_iv,1),1)]\varDep_iv;
varResid_iv = varDep_iv - [varIndep_iv,ones(size(varIndep_iv,1),1)]*varBeta_iv;

varSigmaFull_iv = kron((1/(size(varResid_iv,1)-size(varIndep_iv,2)-1))*(varResid_iv'*varResid_iv),eye(size(varResid_iv,1),size(varResid_iv,1)));
varX_iv = kron(eye(numVars,numVars),[varIndep_iv,ones(size(varIndep_iv,1),1)]);
varVBeta_iv = ((varX_iv'*varX_iv)^(-1))*(varX_iv'*varSigmaFull_iv*varX_iv)*((varX_iv'*varX_iv)^(-1));
for i=1:numVars
    thisFStatR_iv = zeros(lags,numVars*(size(varIndep_iv,2)+1));
    for thislag=1:lags
        thisFStatR_iv(thislag,(i-1)*(size(varIndep_iv,2)+1)+thislag) = 1;
    end
    thisFStat_iv = ((thisFStatR_iv*varBeta_iv(:))'*((thisFStatR_iv*varVBeta_iv*thisFStatR_iv')^(-1))*(thisFStatR_iv*varBeta_iv(:)))/size(thisFStatR_iv,1);
    Fstats_invert(i) = thisFStat_iv;
    Fstats_pval_invert(i) = 1-chi2cdf(thisFStat_iv*size(thisFStatR_iv,1),size(thisFStatR_iv,1));
end

fprintf(strcat('Variable','\t','F stat (IV coeffs = 0)','\t','p value','\n'));
for i=1:numVars
    fprintf(strcat('(',num2str(i),')','\t\t\t',sprintf('%.3f',Fstats_invert(i)),'\t\t\t\t\t',sprintf('%.3f',Fstats_pval_invert(i)),'\n'));
end
fprintf('\n');

%% Summarize properties of expected real rate in VAR

varDepPred = [varIndep,ones(size(varIndep,1),1)]*varBeta;
rDepPred = nan(size(varDepPred,1)-2,3);
rDepPred(:,1) = varDepPred(1:size(varDepPred,1)-2,find(strcmp(rName,otherVars),1)+1);

for horiz=2:3
    varThisDep = varDep(horiz:size(varDep,1),:);
    varThisIndep = varIndep(1:size(varIndep,1)-(horiz-1),:);
    varThisBeta = [varThisIndep,ones(size(varThisIndep,1),1)]\varThisDep;
    varThisDepPred = [varThisIndep,ones(size(varThisIndep,1),1)]*varThisBeta;
    
    rDepPred(:,horiz) = varThisDepPred(1:size(varThisDepPred,1)-(3-horiz),find(strcmp(rName,otherVars),1)+1);
end

rqtr = sum(rDepPred(:,1:3),2);
fprintf(strcat('sigma(4E(real rate)) = ','\t\t\t\t',sprintf('%.3f',4*std(rqtr(1:3:(3*floor((size(rqtr,1)-1)/3)+1)))),'\n'));
fprintf(strcat('rho(4E(real rate)) = ','\t\t\t\t',sprintf('%.3f',corr(rqtr(4:3:(3*floor((size(rqtr,1)-1)/3)+1)),rqtr(1:3:(3*(floor((size(rqtr,1)-1)/3)-1)+1)))),'\n'));
fprintf('\n');

%% Run LP-IV

localIRF = nan(T,numVars);
localIRFSE = nan(T,numVars);
localIRFlCI = nan(T,numVars);
localIRFuCI = nan(T,numVars);

ivIndep = ivData(:,find(strcmp(iv,ivNames),1));
localIndep_iv = lagmatrix(ivData(:,find(strcmp(iv,ivNames),1)),1:lagsIV);
localIndep_var = lagmatrix(varData(mapIVToVar,varInds),1:lagslocal);
localIndep = [localIndep_iv(max(lagsIV,lagslocal)+1:size(localIndep_iv,1),:),...
    localIndep_var(max(lagsIV,lagslocal)+1:size(localIndep_var,1),:)];

% Run first stage
firstDep = varData(mapIVToVar((max(lagsIV,lagslocal)+1):length(mapIVToVar)),varInds(1));
firstIndep = [ivIndep((max(lagsIV,lagslocal)+1):size(ivData,1)),localIndep,ones(size(localIndep,1),1)];
firstBeta = firstIndep\firstDep;
firstPred = firstIndep*firstBeta;
firstResid = firstDep - firstIndep*firstBeta;
firstRobustSE = sqrt(((firstIndep'*firstIndep)^(-1))*...
    (firstIndep'*(size(firstIndep,1)/(size(firstIndep,1)-size(firstIndep,2)))*diag(firstResid.^2)*firstIndep)*...
    ((firstIndep'*firstIndep)^(-1)));
firstFStatR = [1,zeros(1,size(firstIndep,2)-1)];
firstFStat = ((firstFStatR*firstBeta)'*((firstFStatR*(firstRobustSE.^2)*firstFStatR')^(-1))*(firstFStatR*firstBeta))/size(firstFStatR,1);

for i=1:numVars
    for t=0:(T-1)
        thisDep = varData(mapIVToVar((max(lagsIV,lagslocal)+1)+t:length(mapIVToVar)),varInds(i));
        
        % Re-run first stage on the relevant subsample
        thisfirstDep = firstDep(1:size(firstIndep,1)-t);
        thisfirstIndep = firstIndep(1:size(firstIndep,1)-t,:);
        thisfirstBeta = thisfirstIndep\thisfirstDep;
        thisfirstPred = thisfirstIndep*thisfirstBeta;
        thisIndepPred = [thisfirstPred,localIndep(1:size(localIndep,1)-t,:),ones(size(localIndep,1)-t,1)];
        thisIndep = [varData(mapIVToVar((max(lagsIV,lagslocal)+1):length(mapIVToVar)-t),varInds(1)),...
             localIndep(1:size(localIndep,1)-t,:),ones(size(localIndep,1)-t,1)];

        % Run second stage
        localBeta = thisIndepPred\thisDep;
        localResid = thisDep - ...
             thisIndep*localBeta;
        localPZ = thisfirstIndep*((thisfirstIndep'*...
            thisfirstIndep)^(-1))*thisfirstIndep';
        localRobustSE = sqrt(((thisIndep'*localPZ*thisIndep)^(-1))*...
            (thisIndep'*thisfirstIndep*((thisfirstIndep'*thisfirstIndep)^(-1))*...
            thisfirstIndep'*(size(thisIndep,1)/(size(thisIndep,1)-size(thisIndep,2)))*diag(localResid.^2)*thisfirstIndep*...
            ((thisfirstIndep'*thisfirstIndep)^(-1))*thisfirstIndep'*thisIndep)*...
            ((thisIndep'*localPZ*thisIndep)^(-1)));
         % (for a TSLS estimator)

        localIRF(t+1,i) = localBeta(1);
        localIRFSE(t+1,i) = localRobustSE(1);
        localIRFlCI(t+1,i) = localBeta(1) - tinv(uCI/100,size(thisIndep,1)-size(thisIndep,2))*localRobustSE(1);
        localIRFuCI(t+1,i) = localBeta(1) + tinv(uCI/100,size(thisIndep,1)-size(thisIndep,2))*localRobustSE(1);      
    end
end

%% Run recursive VAR with external IV ordered first

rvar1Dep = ivData(lags+1:size(ivData,1),find(strcmp(iv,ivNames),1));
rvar1Indep = lagmatrix(ivData(:,find(strcmp(iv,ivNames),1)),1:lags);
rvarIndep = lagmatrix(varData(:,varInds),1:lags);
rvar1Indep = [rvar1Indep(lags+1:size(rvar1Indep,1),:),rvarIndep(mapIVToVar(lags+1:size(mapIVToVar,1)),:)];
rvar2Dep = varData(mapIVToVar(lags+1:size(mapIVToVar,1)),varInds);
rvar11Indep = lagmatrix(ivData(:,find(strcmp(iv,ivNames),1)),0:lags);
rvar2Indep = [rvar11Indep(lags+1:size(rvar11Indep,1),:),rvarIndep(mapIVToVar(lags+1:size(mapIVToVar,1)),:)];

% Estimate recursive VAR 

rvar1Beta = [rvar1Indep,ones(size(rvar1Indep,1),1)]\rvar1Dep;
rvar1Resid = rvar1Dep - [rvar1Indep,ones(size(rvar1Indep,1),1)]*rvar1Beta;
rvar2Beta = [rvar2Indep,ones(size(rvar2Indep,1),1)]\rvar2Dep;
rvar2Resid = rvar2Dep - [rvar2Indep,ones(size(rvar2Indep,1),1)]*rvar2Beta;

% Stack VAR into a first-order system
% Let rA1, rA2 be matrices of coefficients on non-constants
rA1 = nan(lags+1,lags*(numVars+1));
rA1(1,:) = rvar1Beta(1:(size(rvar1Beta,1)-1))';
rA1(2:(lags+1),:) = [eye(lags,lags),zeros(lags,lags*numVars)];

rA2 = nan(lags*numVars,lags+1+lags*numVars);
rA2(1:numVars,:) = rvar2Beta(1:(size(rvar2Beta,1)-1),:)';
rA2(numVars+1:lags*numVars,:) = [zeros((lags-1)*numVars,lags+1),eye((lags-1)*numVars,(lags-1)*numVars),zeros((lags-1)*numVars,numVars)];

% Generate impulse responses (normalize size such that effect on MP indicator is as in baseline)
rirf = nan(10000,numVars);
rendo1 = zeros(10000,lags+1);
rendo2 = zeros(10000,numVars*lags);
rendo1(1,:) = [sp/rvar2Beta(1,1),zeros(1,lags)];
rendo2(1,:) = (rA2*[rendo1(1,:)';zeros(lags*numVars,1)])';
rirf(1,:) = rendo2(1,1:numVars);
for t=2:10000
	rendo1(t,:) = (rA1*[rendo1(t-1,1:lags)';rendo2(t-1,:)'])';
	rendo2(t,:) = (rA2*[rendo1(t,:)';rendo2(t-1,:)'])';
	rirf(t,:) = rendo2(t,1:numVars);
end

% Decompose contemporaneous effects of MP *loosening* on excess returns 
ruExcessDecompMP = nan(1,4);
ruExcessDecompMP(1) = - rirf(1,find(strcmp(yName,otherVars),1)+1) - ...
    rirf(1,find(strcmp(rName,otherVars),1)+1);    
ruExcessDecompMP(2) = sum((rho.^(1:10000-1)').*rirf(2:10000,find(strcmp(yName,otherVars),1)+1));
ruExcessDecompMP(3) = sum((rho.^(1:10000-1)').*rirf(2:10000,find(strcmp(rName,otherVars),1)+1));
ruExcessDecompMP(4) = ruExcessDecompMP(:,1) - ruExcessDecompMP(:,2) - ruExcessDecompMP(:,3);

%% Run wild bootstrap following Mertens and Ravn (2013) for IRF standard errors and Stock-Watson (18) test statistic 

irfBoots = nan(size(irf,1),size(irf,2),numBoots);
uExcessDecompMPBoots = nan(numBoots,4);
localIRFBoots = nan(size(localIRF,1),size(localIRF,2),numBoots);
rirfBoots = nan(size(irf,1),size(irf,2),numBoots);
rng(0,'twister');
signs = rand(size(varDep,1),numBoots);
signs(signs > 0.5) = 1;
signs(signs <= 0.5) = -1;
for nBoot=1:numBoots
    % Generate data
    endo = zeros(size(varData,1)-lags+1,numVars*lags);
    endo(1,:) = varIndep(1,:);
    u = varResid.*repmat(signs(:,nBoot),1,numVars);
    for t=2:size(endo,1)
        endo(t,:) = (A*(endo(t-1,:)'))' + ...
            [varBeta(numVars*lags+1,:),zeros(1,numVars*(lags-1))] + ...
            [u(t-1,:),zeros(1,numVars*(lags-1))];
    end
    varDataBootAftInit = endo(2:size(endo,1),1:numVars);
    varDataBoot = [varData(1:lags,varInds);varDataBootAftInit];
    ivDataBoot = ivData.*repmat(signs(mapIVToVar-lags,nBoot),1,size(ivData,2));
    
    % Run VAR on simulated data
    varDepBoot = varDataBoot(lags+1:size(varDataBoot,1),:);
    varIndepBoot = lagmatrix(varDataBoot,1:lags);
    varIndepBoot = varIndepBoot(lags+1:size(varIndepBoot,1),:);
    varBetaBoot = [varIndepBoot,ones(size(varIndepBoot,1),1)]\varDepBoot;
    varResidBoot =  varDepBoot - [varIndepBoot,ones(size(varIndepBoot,1),1)]*varBetaBoot;
    varSigmaBoot = (1/(size(ivDataBoot,1)-(lags*numVars+1)))*varResidBoot(mapIVToVar-lags,:)'*varResidBoot(mapIVToVar-lags,:);
    
    % Generate impulse responses and save
    ABoot = nan(lags*numVars,lags*numVars);
    ABoot(1:numVars,:) = varBetaBoot(1:(size(varBetaBoot,1)-1),:)';
    ABoot(numVars+1:lags*numVars,:) = [eye((lags-1)*numVars),zeros((lags-1)*numVars,numVars)];
    [irfBoot,spBoot,sqBoot] = genIRFIV(T,varSigmaBoot,varResidBoot,ivDataBoot,ivNames,mapIVToVar,lags,ABoot,otherVars,iv);
    irfBoots(:,:,nBoot) = irfBoot;
    
    % Decompose contemporaneous effects of MP *loosening* on excess returns
    uExcessDecompMPBoots(nBoot,1) = - irfBoot(1,find(strcmp(yName,otherVars),1)+1) - ...
        irfBoot(1,find(strcmp(rName,otherVars),1)+1);
    uExcessDecompMPBoots(nBoot,2) = sy*rho*ABoot*((eye(numVars*lags)-rho*ABoot)^(-1))*[spBoot;sqBoot;zeros((lags-1)*numVars,1)];
    uExcessDecompMPBoots(nBoot,3) = sr*rho*ABoot*((eye(numVars*lags)-rho*ABoot)^(-1))*[spBoot;sqBoot;zeros((lags-1)*numVars,1)];
    uExcessDecompMPBoots(nBoot,4) = uExcessDecompMPBoots(nBoot,1) - uExcessDecompMPBoots(nBoot,2) - uExcessDecompMPBoots(nBoot,3);
    
    % Run local projection
    ivIndepBoot = ivDataBoot(:,find(strcmp(iv,ivNames),1));
    localIndep_ivBoot = lagmatrix(ivDataBoot(:,find(strcmp(iv,ivNames),1)),1:lagsIV);
    localIndep_varBoot = lagmatrix(varDataBoot(mapIVToVar,:),1:lagslocal);
    localIndepBoot = [localIndep_ivBoot(max(lagsIV,lagslocal)+1:size(localIndep_ivBoot,1),:),...
        localIndep_varBoot(max(lagsIV,lagslocal)+1:size(localIndep_varBoot,1),:)];
    
    firstDepBoot = varDataBoot(mapIVToVar((max(lagsIV,lagslocal)+1):length(mapIVToVar)),1);
    firstIndepBoot = [ivIndepBoot((max(lagsIV,lagslocal)+1):size(ivDataBoot,1)),localIndepBoot,ones(size(localIndepBoot,1),1)];

    for i=1:numVars
        for t=1:12:(T-1)
            thisDepBoot = varDataBoot(mapIVToVar((max(lagsIV,lagslocal)+1)+t:length(mapIVToVar)),i);

            % Re-run first stage on the relevant subsample
            thisfirstDepBoot = firstDepBoot(1:size(firstIndepBoot,1)-t);
            thisfirstIndepBoot = firstIndepBoot(1:size(firstIndepBoot,1)-t,:);
            thisfirstBetaBoot = thisfirstIndepBoot\thisfirstDepBoot;
            thisfirstPredBoot = thisfirstIndepBoot*thisfirstBetaBoot;
            thisIndepPredBoot = [thisfirstPredBoot,localIndepBoot(1:size(localIndepBoot,1)-t,:),ones(size(localIndepBoot,1)-t,1)];
            thisIndepBoot = [varDataBoot(mapIVToVar((max(lagsIV,lagslocal)+1):length(mapIVToVar)-t),1),...
                 localIndepBoot(1:size(localIndep,1)-t,:),ones(size(localIndep,1)-t,1)];

            % Run second stage
            localBetaBoot = thisIndepPredBoot\thisDepBoot;
            localIRFBoots(t+1,i,nBoot) = localBetaBoot(1);     
        end
    end    
    
    % Generate data for recursive VAR
    rendo1Boot = zeros(size(rvar1Dep,1)+1,lags+1);
    rendo2Boot = zeros(size(rvar1Dep,1)+1,numVars*lags);
    rendo1Boot(1,1:lags) = rvar2Indep(1,2:(lags+1));
    rendo2Boot(1,:) = rvar2Indep(1,lags+2:size(rvar2Indep,2));
    ru1 = rvar1Resid.*signs(mapIVToVar(lags+1:size(mapIVToVar,1))-lags,nBoot);
    ru2 = rvar2Resid.*repmat(signs(mapIVToVar(lags+1:size(mapIVToVar,1))-lags,nBoot),1,numVars);
    for t=2:size(rendo1Boot,1)
        rendo1Boot(t,:) = (rA1*[rendo1Boot(t-1,1:lags)';rendo2Boot(t-1,:)'])' + ...
            [rvar1Beta((numVars+1)*lags+1),zeros(1,lags)] + ...
            [ru1(t-1,:),zeros(1,lags)];
        rendo2Boot(t,:) = (rA2*[rendo1Boot(t,:)';rendo2Boot(t-1,:)'])' + ...
            [rvar2Beta(numVars*lags+1,:),zeros(1,numVars*(lags-1))] + ...
            [ru2(t-1,:),zeros(1,numVars*(lags-1))];
    end
    rivDataBootAftInit = rendo1Boot(2:size(rendo1Boot,1),1);
    rivDataBoot = [ivData(1:lags,find(strcmp(iv,ivNames),1));rivDataBootAftInit];
    rvarDataBootAftInit = rendo2Boot(2:size(rendo2Boot,1),1:numVars);
    rvarDataBoot = [varData(mapIVToVar(1:lags),varInds);rvarDataBootAftInit];    
    
    % Run recursive VAR on simulated data
    rvar1DepBoot = rivDataBoot(lags+1:size(rivDataBoot,1));
    rvar1IndepBoot = lagmatrix(rivDataBoot,1:lags);
    rvarIndepBoot = lagmatrix(rvarDataBoot,1:lags);
    rvar1IndepBoot = [rvar1IndepBoot(lags+1:size(rvar1IndepBoot,1),:),rvarIndepBoot(lags+1:size(rvarIndepBoot,1),:)];
    rvar2DepBoot = rvarDataBoot(lags+1:size(rvarDataBoot,1),:);
    rvar11IndepBoot = lagmatrix(rivDataBoot,0:lags);
    rvar2IndepBoot = [rvar11IndepBoot(lags+1:size(rvar11IndepBoot,1),:),rvarIndepBoot(lags+1:size(rvarIndepBoot,1),:)];  
    
    rvar1BetaBoot = [rvar1IndepBoot,ones(size(rvar1IndepBoot,1),1)]\rvar1DepBoot;
    rvar2BetaBoot = [rvar2IndepBoot,ones(size(rvar2IndepBoot,1),1)]\rvar2DepBoot;
       
    % Generate impulse responses and save
    rA1Boot = nan(lags+1,lags*(numVars+1));
    rA1Boot(1,:) = rvar1BetaBoot(1:(size(rvar1BetaBoot,1)-1))';
    rA1Boot(2:(lags+1),:) = [eye(lags,lags),zeros(lags,lags*numVars)];

    rA2Boot = nan(lags*numVars,lags+1+lags*numVars);
    rA2Boot(1:numVars,:) = rvar2BetaBoot(1:(size(rvar2BetaBoot,1)-1),:)';
    rA2Boot(numVars+1:lags*numVars,:) = [zeros((lags-1)*numVars,lags+1),eye((lags-1)*numVars,(lags-1)*numVars),zeros((lags-1)*numVars,numVars)];

    rirfBoot = nan(size(irf,1),numVars);
    rendo1Boot = zeros(size(irf,1),lags+1);
    rendo2Boot = zeros(size(irf,1),numVars*lags);
    rendo1Boot(1,:) = [sp/rvar2BetaBoot(1,1),zeros(1,lags)]; 
     % always normalize to have same effect on 1-year Treasury as baseline
     % (since only relative IRFs are estimated)
    rendo2Boot(1,:) = (rA2Boot*[rendo1Boot(1,:)';zeros(lags*numVars,1)])';
    rirfBoot(1,:) = rendo2Boot(1,1:numVars);
    for t=2:size(irf,1)
        rendo1Boot(t,:) = (rA1Boot*[rendo1Boot(t-1,1:lags)';rendo2Boot(t-1,:)'])';
        rendo2Boot(t,:) = (rA2Boot*[rendo1Boot(t,:)';rendo2Boot(t-1,:)'])';
        rirfBoot(t,:) = rendo2Boot(t,1:numVars);
    end    
    rirfBoots(:,:,nBoot) = rirfBoot;
    
    if mod(nBoot,100) == 0
        disp(num2str(nBoot));
    end
end
irflCI = nan(size(irf));
irfuCI = nan(size(irf));
for t=1:size(irf,1)
    irflCI(t,:) = prctile(squeeze(irfBoots(t,:,:))',lCI);
    irfuCI(t,:) = prctile(squeeze(irfBoots(t,:,:))',uCI);
end
uExcessDecompMPlCI = prctile(uExcessDecompMPBoots,lCI);
uExcessDecompMPuCI = prctile(uExcessDecompMPBoots,uCI);
rirflCI = nan(size(irf));
rirfuCI = nan(size(irf));
for t=1:size(irf,1)
    rirflCI(t,:) = prctile(squeeze(rirfBoots(t,:,:))',lCI);
    rirfuCI(t,:) = prctile(squeeze(rirfBoots(t,:,:))',uCI);
end

fprintf(strcat('Effect of MP surprise on S&P 500 return:','\t\t\t\t',...
    sprintf('%.3f',uExcessDecompMP(1)),'\t','[',sprintf('%.3f',uExcessDecompMPlCI(1)),...
    ',',sprintf('%.3f',uExcessDecompMPuCI(1)),']','\n'));    
fprintf(strcat('-Effect of MP surprise on future excess returns:','\t\t',...
    sprintf('%.3f',uExcessDecompMP(2)),'\t','[',sprintf('%.3f',uExcessDecompMPlCI(2)),...
    ',',sprintf('%.3f',uExcessDecompMPuCI(2)),']','\n'));
fprintf(strcat('-Effect of MP surprise on future real interest rates:','\t',...
    sprintf('%.3f',uExcessDecompMP(3)),'\t','[',sprintf('%.3f',uExcessDecompMPlCI(3)),...
    ',',sprintf('%.3f',uExcessDecompMPuCI(3)),']','\n'));    
fprintf(strcat('Effect of MP surprise on dividends:','\t\t\t\t\t\t',...
    sprintf('%.3f',uExcessDecompMP(4)),'\t','[',sprintf('%.3f',uExcessDecompMPlCI(4)),...
    ',',sprintf('%.3f',uExcessDecompMPuCI(4)),']','\n'));

uExcessDecompMPBoots_sh = nan(size(uExcessDecompMPBoots,1),3);
uExcessDecompMPBoots_sh(:,1) = uExcessDecompMPBoots(:,2)./uExcessDecompMPBoots(:,1);
uExcessDecompMPBoots_sh(:,2) = uExcessDecompMPBoots(:,3)./uExcessDecompMPBoots(:,1);
uExcessDecompMPBoots_sh(:,3) = uExcessDecompMPBoots(:,4)./uExcessDecompMPBoots(:,1);
uExcessDecompMPlCI_sh = prctile(uExcessDecompMPBoots_sh,lCI);
uExcessDecompMPuCI_sh = prctile(uExcessDecompMPBoots_sh,uCI);

fprintf(strcat('As a fraction of current S&P 500 return:','\n'));    
fprintf(strcat('-Effect of MP surprise on future excess returns:','\t\t',...
sprintf('%.3f',uExcessDecompMP(2)/uExcessDecompMP(1)),'\t','[',sprintf('%.3f',uExcessDecompMPlCI_sh(1)),...
',',sprintf('%.3f',uExcessDecompMPuCI_sh(1)),']','\n'));
fprintf(strcat('-Effect of MP surprise on future real interest rates:','\t',...
    sprintf('%.3f',uExcessDecompMP(3)/uExcessDecompMP(1)),'\t','[',sprintf('%.3f',uExcessDecompMPlCI_sh(2)),...
    ',',sprintf('%.3f',uExcessDecompMPuCI_sh(2)),']','\n'));    
fprintf(strcat('Effect of MP surprise on dividends:','\t\t\t\t\t\t',...
    sprintf('%.3f',uExcessDecompMP(4)/uExcessDecompMP(1)),'\t','[',sprintf('%.3f',uExcessDecompMPlCI_sh(3)),...
    ',',sprintf('%.3f',uExcessDecompMPuCI_sh(3)),']','\n'));
fprintf('\n');

% Compute Stock-Watson (2018) test stats
fprintf(strcat('Variable','\t','Test stat (SW 2018)','\t\t','p value','\n'));
for i=1:numVars
    % Estimate scaled variance matrix (1/T)*V 
    thisSWDiffBoot = nan(floor(T/12),numBoots);
    for t=1:12:(T-1)
        thisSWDiffBoot(floor(t/12)+1,:) = reshape(irfBoots(t+1,i,:)./irfBoots(1,1,:),1,numBoots) - ...
            reshape(localIRFBoots(t+1,i,:),1,numBoots);
         % divide by irfBoots(1,1,:) so that horizon 0 effect on monetary 
         %  policy indicator is 1 across all samples
    end
    thisSWDiffBootMean = mean(thisSWDiffBoot,2);
    thisSWscV = (1/(numBoots-1))*(thisSWDiffBoot-repmat(thisSWDiffBootMean,1,numBoots))*...
        (thisSWDiffBoot-repmat(thisSWDiffBootMean,1,numBoots))';
    
    % Construct test statistic
    thisSWDiff = nan(floor(T/12),1);
    for t=1:12:(T-1)
        thisSWDiff(floor(t/12)+1) = irf(t+1,i)./irf(1,1) - localIRF(t+1,i);
    end
    thisSWTestStat = thisSWDiff'*(thisSWscV^(-1))*thisSWDiff;
    
    fprintf(strcat('(',num2str(i),')','\t\t\t',sprintf('%.3f',thisSWTestStat),...
        '\t\t\t\t\t',sprintf('%.3f',1-chi2cdf(thisSWTestStat,floor(T/12))),'\n'));
end
fprintf('\n');

%% Compute relevant point estimates for alternative specifications (only for baseline)

if ~includeRelFund
    % I save and print [initial MP indicator response, initial S&P 500 return response, share excess return component]
    %  for an unexpected monetary *loosening*

    fprintf(['mpind_0','\t\t','e_0','\t\t\t','Sh e','\n']);    

    % Baseline
    fprintf(['Baseline','\n']);
    fprintf([sprintf('%.3f',-irf(1,1)),'\t\t',...
        sprintf('%.3f',uExcessDecompMP(1)),'\t\t',...
        sprintf('%.3f',uExcessDecompMP(2)/uExcessDecompMP(1)),'\n']);
    
    % Number of lags
    fprintf(['Lags','\n']);
    robustLags = [4,5,7,8];
    for i=1:length(robustLags)
        [robustirf,robustuExcessDecompMP] = runRobustness(numVars,numOtherVars,robustLags(i),...
            varData,ivData,varNames,ivNames,mpInd,otherVars,yName,rName,iv,...
            T,mapIVToVar,rho);
        fprintf([sprintf('%.3f',-robustirf(1,1)),'\t\t',...
            sprintf('%.3f',robustuExcessDecompMP(1)),'\t\t',...
            sprintf('%.3f',robustuExcessDecompMP(2)/robustuExcessDecompMP(1)),'\n']);
    end

    % VAR sample periods
    fprintf(['VAR sample','\n']);
    [robustirf,robustuExcessDecompMP] = runRobustness(numVars,numOtherVars,lags,...
        varData([[mapIVToVar(1)-lags:mapIVToVar(1)-1]';mapIVToVar],:),ivData,varNames,ivNames,mpInd,otherVars,yName,rName,iv,...
        T,mapIVToVar-mapIVToVar(1)+lags+1,rho);
    fprintf([sprintf('%.3f',-robustirf(1,1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(2)/robustuExcessDecompMP(1)),'\n']);

    % IV sample periods
    fprintf(['IV sample','\n']);
    [robustirf,robustuExcessDecompMP] = runRobustness(numVars,numOtherVars,lags,...
        varData,ivData(1:floor(size(ivData,1)/2),:),varNames,ivNames,mpInd,otherVars,yName,rName,iv,...
        T,mapIVToVar(1:floor(size(ivData,1)/2)),rho);
    fprintf([sprintf('%.3f',-robustirf(1,1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(2)/robustuExcessDecompMP(1)),'\n']);
    [robustirf,robustuExcessDecompMP] = runRobustness(numVars,numOtherVars,lags,...
        varData,ivData(floor(size(ivData,1)/2)+1:size(ivData,1),:),varNames,ivNames,mpInd,otherVars,yName,rName,iv,...
        T,mapIVToVar(floor(size(ivData,1)/2)+1:size(ivData,1)),rho);
    fprintf([sprintf('%.3f',-robustirf(1,1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(2)/robustuExcessDecompMP(1)),'\n']);

    % Variables included in VAR
    fprintf(['Additional variables','\n']);
    newVariables = {'ebp','mortg_spread_m','cp3m_spread_m','cm5yr','cm10yr','termspread','relbill','dyield3mo'};
    for i=1:length(newVariables)
        [robustirf,robustuExcessDecompMP] = runRobustness(numVars+1,numOtherVars+1,lags,...
            varData,ivData,varNames,ivNames,mpInd,[otherVars,newVariables{i}],yName,rName,iv,...
            T,mapIVToVar,rho);
        fprintf([sprintf('%.3f',-robustirf(1,1)),'\t\t',...
            sprintf('%.3f',robustuExcessDecompMP(1)),'\t\t',...
            sprintf('%.3f',robustuExcessDecompMP(2)/robustuExcessDecompMP(1)),'\n']);
    end

    % Using 'ff4_tc' as the IV
    fprintf(['Using ff4_tc as the IV','\n']);
    [robustirf,robustuExcessDecompMP] = runRobustness(numVars,numOtherVars,lags,...
        varData,ivData,varNames,ivNames,mpInd,otherVars,yName,rName,'ff4_tc',...
        T,mapIVToVar,rho);
    fprintf([sprintf('%.3f',-robustirf(1,1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(1)),'\t\t',...
        sprintf('%.3f',robustuExcessDecompMP(2)/robustuExcessDecompMP(1)),'\n']);

    % Recursive VAR with IV ordered first (scaled so that horizon 0 effect on
    %  monetary policy indicator is same as SVAR-IV)
    fprintf(['Recursive VAR with IV ordered first','\n']);
    fprintf([sprintf('%.3f',-rirf(1,1)),'\t\t',...
        sprintf('%.3f',ruExcessDecompMP(1)),'\t\t',...
        sprintf('%.3f',ruExcessDecompMP(2)/ruExcessDecompMP(1)),'\n']);
end

%% Plot impulse responses

diary off;
Interpreter = 'latex';
fontsize = 14;
darkblue    = [8,69,148]./255;
lightblue   = [107,174,214]./255;
verylightblue = [158,202,225]./255;
color_list = {darkblue, verylightblue};
color_list_bw = {rgb2gray(darkblue), rgb2gray(verylightblue)};
   
for bw=0:1
    if bw
        this_color_list = color_list_bw;
    else
        this_color_list = color_list;
    end
    
    if ~includeRelFund % then plot all impulse responses
        svarfig = figure;
        for i=1:numVars
            subplot(rowsToPlot,colsToPlot,i);
            yplot = signToPlot*[irflCI(1:T,i)', flipud(irfuCI(1:T,i))'];
            plot(0:(T-1),zeros(1,T),'-k');
            hold on;
            fill([0:(T-1),(T-1):-1:0],yplot,1,'facecolor', this_color_list{2}, 'edgecolor', 'none','facealpha',1.0);
            plot(0:(T-1),zeros(1,T),'-k');
            plot(0:(T-1),signToPlot*irf(1:T,i)','LineWidth',2,'Color',this_color_list{1},'Linestyle','-');
            if strcmp(varSeriesToPrint{i},'Dividend/price')
                ylim([-0.02,0.01]);
            end
            hold off;
            xlim([0,(T-1)]);
            ylabel('pp','Fontsize',fontsize,'Interpreter',Interpreter); 
            title(varSeriesToPrint{i},'Fontsize',fontsize,'Interpreter',Interpreter);
            ax2 = gca;
            ax2.FontSize = fontsize;
            ax2.TickLabelInterpreter = Interpreter;
        end
        set(gcf,'PaperType','usletter','PaperOrientation','portrait','PaperPosition',[0 0 8.5 5.0]);
        if bw
            print(gcf, ['output/',fileName,'_bw'], '-depsc', '-r600','-loose'); 
        else
            print(gcf, ['output/',fileName], '-depsc', '-r600','-loose'); 
        end
    else % then just plot relative wealth responses (assumed to be last two variables in VAR)
        svarfig_relwealth = figure;
        for i=numVars-1:numVars
            subplot(1,2,(i-(numVars-2)));
            yplot = signToPlot*[irflCI(1:T,i)', flipud(irfuCI(1:T,i))'];
            plot(0:(T-1),zeros(1,T),'-k');
            hold on;
            fill([0:(T-1),(T-1):-1:0],yplot,1,'facecolor', this_color_list{2}, 'edgecolor', 'none','facealpha',1.0);
            plot(0:(T-1),zeros(1,T),'-k');   
            plot(0:(T-1),signToPlot*irf(1:T,i)','LineWidth',2,'Color',this_color_list{1},'Linestyle','-');    
            hold off;
            xlim([0,(T-1)]);
            ylabel('pp','Fontsize',fontsize,'Interpreter',Interpreter);
            title(varSeriesToPrint{i},'Fontsize',fontsize,'Interpreter',Interpreter);
            ax2 = gca;
            ax2.FontSize = fontsize;
            ax2.TickLabelInterpreter = Interpreter;
            ylim(ax2,[-2,2]);
        end
        set(gcf,'PaperType','usletter','PaperOrientation','portrait','PaperPosition',[0 0 5.67 2.5]);
        if bw
            print(gcf, ['output/',fileName,'_bw'], '-depsc', '-r600','-loose');   
        else
            print(gcf, ['output/',fileName], '-depsc', '-r600','-loose');
        end
    end
end