* computeMoments2016.do
* Summarizes distribution of household wealth and portfolios in 2016 SCF

#delimit;

clear all;
set maxvar 10000;

timer clear 1;
timer on 1;

local dir "C:\...\Empirical Micro";

local al_exvh_pct = 60;
 * defines percentile of wealth to labor income distribution to cut at;
local rsh_exvh_pct = 90; 
 * defines percentile of portfolio share distribution to cut at, among high wealth to labor income group;
scalar indHet = 1;
 * if = 1, account for idiosyncratic heterogeneity in leverage (default: 1);
local numDraws = 20;
 * number of idiosyncratic leverage draws for each household;
 * (only relevant if indHet = 1);
local targetVolAgg = 1.81/5.8;
 * target ratio of standard deviation gross leverage to aggregate firm leverage (matching estimates of Bach-Calvet-Sodini (2018));
 * (only relevant if indHet = 1);
local gammaShape = 0.95463465;
 * initial value of shape parameter for Gamma distribution;
 
matrix rsh_exvh_pct_vec = (50\51\52\53\54\55\56\57\58\59\
				  60\61\62\63\64\65\66\67\68\69\
				  70\71\72\73\74\75\76\77\78\79\
				  80\81\82\83\84\85\86\87\88\89\
				  90\91\92\93\94\95\96\97\98\99);
 * defines percentile of portfolio share distribution to cut at, among high wealth to labor income group (for sensitivity analysis);
 
local rsh_exvh_lbound = -10000;
 * ignore observations with rsh_exvh < lbound;
local rsh_exvh_ubound = 10000;
 * ignore observations with rsh_exvh > ubound;

local firmnetlev = 1.6;
* source: Q2 2019 release of FA for 2016; 
local actbusnetlev = 1.1;
* source: Q2 2019 release of FA for 2016;

use "`dir'\data\scf2016.dta", replace;
	
* Draw individual leverage to target targetVolAgg (if indlev = 1);
gen indfirmnetlev = `firmnetlev';
gen indactbusnetlev = `actbusnetlev';
	
replace wgt = wgt/`numDraws';
expand `numDraws', gen(newFlag);

set seed 1000;
gen udraw = runiform();
local volAgg = `targetVolAgg';
scalar continueDecomp = 1;
while continueDecomp {;
	* Set gammaShape and draw individual leverage;
	if indHet {;
		local gammaShape = `gammaShape' - 0.5*(1-`volAgg'/`targetVolAgg');
		local firmScale = `firmnetlev'/`gammaShape';
		local actbusScale = `actbusnetlev'/`gammaShape';
		qui replace indfirmnetlev = `firmScale'*invgammap(`gammaShape',udraw);
		qui replace indactbusnetlev = `actbusScale'*invgammap(`gammaShape',udraw);
	};
 	
	* Compute positions in capital and bonds for each household;
	gen equity_K = equity*indfirmnetlev`year';
	gen equity_B = equity - equity_K;
	
	qui gen actbus_K = actbus*indactbusnetlev;
	qui gen actbus_B = actbus*(1-indactbusnetlev);
	
	qui gen nonactbus_K = nonactbus*indactbusnetlev;
	qui gen nonactbus_B = nonactbus*(1-indactbusnetlev);
	
	qui gen vehic_B = 0;
	qui gen vehic_K = vehic;
	
	qui gen houses_B = 0;
	qui gen houses_K = houses;	
	
	qui gen oresre_B = 0;
	qui gen oresre_K = oresre;
	
	qui gen nnresre_B = 0;
	qui gen nnresre_K = nnresre;
	
	qui gen othnfin_B = 0;
	qui gen othnfin_K = othnfin;	

	qui gen db_B = db;
	qui gen db_K = 0;	
	
	* Compute total wealth, capital, and bonds for each household;
	gen K = equity_K + actbus_K + nonactbus_K + vehic_K + houses_K + oresre_K + nnresre_K + othnfin_K;
	gen A = networth + db;
	gen B = A - K;		
	
	* Compute total *gross* wealth for each household;
	gen A_gross = A + mrthel + resdbt + othloc + ccbal + install + odebt;	
	
	* Compute standard deviation of leverage, relative to leverage of aggregate business sector;	
	qui gen rsh_gross = K/A_gross;
	qui tabstat rsh_gross [aw=wgt], statistics(sd) save;
	local volAgg = el(r(StatTotal),1,1)/`firmnetlev';
	di `gammaShape';
	di `volAgg';

	if abs(1-`volAgg'/`targetVolAgg') > 0.0001 & indHet {;
		drop *_B *_K A B K *_gross;
	};
	else {;
		scalar continueDecomp = 0;
	};
};

* Compute total wealth, capital, and bonds for each household, excluding vehicles and primary residence;
gen K_exvh = equity_K + actbus_K + nonactbus_K + 
	oresre_K + nnresre_K + othnfin_K;
gen A_exvh = (networth - (vehic + houses) + mrthel + VEH_INST) + db;
gen B_exvh = A_exvh - K_exvh;

* Generate portfolio shares and wealth/labor measures;
foreach theend in "" "_exvh" {;
	gen rsh`theend' = K`theend' / A`theend';
	gen al`theend' = A`theend' / wageinc;
	* replace with large number if wageinc = 0;
	replace al`theend' = 1000000000 if wageinc == 0;
	gen la`theend' = wageinc / A`theend';
};

* Define first group;
_pctile al_exvh [aw=wgt], percentiles(`al_exvh_pct');
scalar al_exvh_bound = r(r1);
gen al_exvh_group = 0 if al_exvh < al_exvh_bound;
replace al_exvh_group = 1 if al_exvh >= al_exvh_bound;

* Within high group, define groups of portfolio share distribution;
_pctile rsh_exvh if al_exvh_group == 1 [aw=wgt], percentiles(`rsh_exvh_pct');
scalar rsh_exvh_bound = r(r1);
gen rsh_exvh_group = 0 if rsh_exvh < rsh_exvh_bound & rsh_exvh >= `rsh_exvh_lbound';
replace rsh_exvh_group = 1 if rsh_exvh >= rsh_exvh_bound & rsh_exvh <= `rsh_exvh_ubound';

* Define composite groups (0 = low wealth to labor income, 1 = high wealth to labor income, low portfolio share, 2 = high wealth to labor income, high portfolio share);
gen group = 0 if al_exvh_group == 0;
replace group = 1 + rsh_exvh_group if al_exvh_group == 1 & rsh_exvh_group < .;

* Summarize characteristics of groups;
matrix summStats = (0,0,0\0,0,0\0,0,0\0,0,0\0,0,0);
tabstat wgt, by(group) statistics(sum) format(%9.3f) save;
matrix summStats[1,1] = el(r(Stat1),1,1)/el(r(StatTotal),1,1);
matrix summStats[1,2] = el(r(Stat2),1,1)/el(r(StatTotal),1,1);
matrix summStats[1,3] = el(r(Stat3),1,1)/el(r(StatTotal),1,1);

tabstat wageinc [aw=wgt], by(group) statistics(sum) format(%9.3f) save;
matrix summStats[2,1] = el(r(Stat1),1,1)/el(r(StatTotal),1,1);
matrix summStats[2,2] = el(r(Stat2),1,1)/el(r(StatTotal),1,1);
matrix summStats[2,3] = el(r(Stat3),1,1)/el(r(StatTotal),1,1);

tabstat A [aw=wgt], by(group) statistics(sum) format(%9.3f) save;
matrix summStats[3,1] = el(r(Stat1),1,1)/el(r(StatTotal),1,1);
scalar total1 = el(r(Stat1),1,1);
matrix summStats[3,2] = el(r(Stat2),1,1)/el(r(StatTotal),1,1);
scalar total2 = el(r(Stat2),1,1);
matrix summStats[3,3] = el(r(Stat3),1,1)/el(r(StatTotal),1,1);	
scalar total3 = el(r(Stat3),1,1);

tabstat K [aw=wgt], by(group) statistics(sum) format(%9.3f) save;
matrix summStats[4,1] = el(r(Stat1),1,1)/total1;
scalar totalK1 = el(r(Stat1),1,1);
matrix summStats[4,2] = el(r(Stat2),1,1)/total2;
scalar totalK2 = el(r(Stat2),1,1);
matrix summStats[4,3] = el(r(Stat3),1,1)/total3;
scalar totalK3 = el(r(Stat3),1,1);

tabstat K_exvh [aw=wgt], by(group) statistics(sum) format(%9.3f) save;
matrix summStats[5,1] = 1-el(r(Stat1),1,1)/totalK1;
matrix summStats[5,2] = 1-el(r(Stat2),1,1)/totalK2;
matrix summStats[5,3] = 1-el(r(Stat3),1,1)/totalK3;	

matrix rownames summStats = "Frac households" "Frac wageinc"
	"Frac A" "K/A" "(Veh+hous)/K";
matrix colnames summStats = "c" "b" "a";

putexcel set "`dir'\output\heterogeneity2016.xlsx", replace;
putexcel A1 = matrix(summStats), names;

* Iterate over alternative cutoffs, computing key moments and saving;
local n_ind = rowsof(rsh_exvh_pct_vec);
matrix fracHHStats = J(`n_ind',3,0);
matrix fracWlStats = J(`n_ind',3,0);
matrix fracAStats = J(`n_ind',3,0);
matrix fracKAStats = J(`n_ind',3,0);
* cols of each of the above matrices are ordered (group 0, group 1, group 2);
* rows of each of the above matrices are ordered by each element of rsh_exvh_pct_vec;
	
forvalues ind = 1(1)`n_ind' {;			
	* Within high group, define groups of portfolio share distribution;
	local this_rsh_exvh_pct = el(rsh_exvh_pct_vec,`ind',1);
	_pctile rsh_exvh if al_exvh_group == 1 [aw=wgt], percentiles(`this_rsh_exvh_pct');
	scalar temp_rsh_exvh_bound = r(r1);
	gen rsh_exvh_temp_group = 0 if rsh_exvh < temp_rsh_exvh_bound & rsh_exvh >= `rsh_exvh_lbound';
	replace rsh_exvh_temp_group = 1 if rsh_exvh >= temp_rsh_exvh_bound & rsh_exvh <= `rsh_exvh_ubound';

	* Define composite groups (0 = low group1, 1 = high group1, low portfolio share, 2 = high group1, high portfolio share);
	gen temp_group = 0 if al_exvh_group == 0;
	replace temp_group = 1 + rsh_exvh_temp_group if al_exvh_group == 1 & rsh_exvh_temp_group < .;
	
	* Compute key moments and save;
	qui tabstat wgt, by(temp_group) statistics(sum) format(%9.3f) save;
	matrix fracHHStats[`ind',1] = el(r(Stat1),1,1)/el(r(StatTotal),1,1);
	matrix fracHHStats[`ind',2] = el(r(Stat2),1,1)/el(r(StatTotal),1,1);
	matrix fracHHStats[`ind',3] = el(r(Stat3),1,1)/el(r(StatTotal),1,1);
	
	qui tabstat wageinc [aw=wgt], by(temp_group) statistics(sum) format(%9.3f) save;
	matrix fracWlStats[`ind',1] = el(r(Stat1),1,1)/el(r(StatTotal),1,1);
	matrix fracWlStats[`ind',2] = el(r(Stat2),1,1)/el(r(StatTotal),1,1);
	matrix fracWlStats[`ind',3] = el(r(Stat3),1,1)/el(r(StatTotal),1,1);
	
	qui tabstat A [aw=wgt], by(temp_group) statistics(sum) format(%9.3f) save;
	matrix fracAStats[`ind',1] = el(r(Stat1),1,1)/el(r(StatTotal),1,1);
	scalar total1 = el(r(Stat1),1,1);
	matrix fracAStats[`ind',2] = el(r(Stat2),1,1)/el(r(StatTotal),1,1);
	scalar total2 = el(r(Stat2),1,1);
	matrix fracAStats[`ind',3] = el(r(Stat3),1,1)/el(r(StatTotal),1,1);
	scalar total3 = el(r(Stat3),1,1);
	
	qui tabstat K [aw=wgt], by(temp_group) statistics(sum) format(%9.3f) save;
	matrix fracKAStats[`ind',1] = el(r(Stat1),1,1)/total1;
	matrix fracKAStats[`ind',2] = el(r(Stat2),1,1)/total2;
	matrix fracKAStats[`ind',3] = el(r(Stat3),1,1)/total3;	

	drop rsh_exvh_temp_group temp_group;			
};

svmat fracHHStats, names("fracHH_");
svmat fracWlStats, names("fracWl_");
svmat fracAStats, names("fracA_");
svmat fracKAStats, names("fracKA_");
svmat rsh_exvh_pct_vec;
* Export for plotting in Matlab for paper;
local sensitivityVars = "rsh_exvh_pct_vec";
	forvalues thisgroup = 1(1)3 {;
		local groupnum = `thisgroup'-1;		
		local sensitivityVars = "`sensitivityVars' fracHH_`thisgroup' fracWl_`thisgroup' fracA_`thisgroup' fracKA_`thisgroup'";
	};
export delimited `sensitivityVars' using "`dir'\output\sensitivity2016.csv", replace;

* Project group indicators on observables;
* Focus on the original sample of households (newFlag == 0) to obtain correct SEs;
gen group1 = 1 if group == 1;
replace group1 = 0 if group == 0 | group == 2;
gen group2 = 1 if group == 2;
replace group2 = 0 if group == 0 | group == 1;
drop if newFlag == 1;
replace wgt = wgt*`numDraws';
gen retired = 1 if lifecl == 6;
replace retired = 0 if lifecl >= 1 & lifecl <= 5;

gen rep = Y1 - YY1*10;
gen wgt0 = wgt*5; 
 * (restores original weight X42001 = wgt*5);
 
scalar depCounter = 1;
matrix obsStats = (0,0\0,0\0,0\0,0\0,0\0,0);
foreach depVar in hbus retired {;
	scfcombo `depVar' group2`theend' group1`theend' [aw=wgt0], command(regress) reps(200) imps(5);
	matrix obsStats[1,depCounter] = el(r(table),1,1);
	matrix obsStats[2,depCounter] = el(r(table),2,1);
	matrix obsStats[3,depCounter] = el(r(table),1,2);
	matrix obsStats[4,depCounter] = el(r(table),2,2);
	qui regress `depVar' group1`theend' group2`theend' [aw=wgt0] if rep==1;
	matrix obsStats[5,depCounter] = e(N);
	matrix obsStats[6,depCounter] = e(r2_a);
	scalar depCounter = depCounter + 1;
};
matrix rownames obsStats = "group = a" "se" "group = b" "se" "N" "Adj R^2";
matrix colnames obsStats = "hbus" "retired";

putexcel set "`dir'\output\hbusretired2016.xlsx", replace;
putexcel A1 = matrix(obsStats), names;

timer off 1;
timer list 1;
